@@ -1,4 +1,6 @@
-from breathe.parser import DoxygenCompoundParser
+import sphinx
+
+from breathe.parser import compound, compoundsuper, DoxygenCompoundParser
 from breathe.project import ProjectInfo
 from breathe.renderer import RenderContext
 from breathe.renderer.filter import Filter
@@ -26,7 +28,6 @@ except ImportError:
     cs = None
 
 import re
-import six
 import textwrap
 from typing import Callable, cast, Dict, List, Optional, Tuple, Type, Union  # noqa
 
@@ -366,7 +367,7 @@ def get_param_decl(param):
         if node is not None:
             for p in node.content_:
                 value = p.value
-                if not isinstance(value, six.text_type):
+                if not isinstance(value, str):
                     value = value.valueOf_
                 result.append(value)
         return ' '.join(result)
@@ -555,7 +556,7 @@ class SphinxRenderer:
         node = node_stack[0]
         # An enumvalue node doesn't have location, so use its parent node for detecting
         # the domain instead.
-        if isinstance(node, six.string_types) or node.node_type == "enumvalue":
+        if isinstance(node, str) or node.node_type == "enumvalue":
             node = node_stack[1]
         filename = get_filename(node)
         if not filename and node.node_type == "compound":
@@ -667,10 +668,23 @@ class SphinxRenderer:
         assert declarator is not None
         if display_obj_type is not None:
             n = declarator[0]
-            assert isinstance(n, addnodes.desc_annotation)
-            assert n.astext()[-1] == " "
-            txt = display_obj_type + ' '
-            declarator[0] = addnodes.desc_annotation(txt, txt)
+            newStyle = True
+            # the new style was introduced in Sphinx v4
+            if sphinx.version_info[0] < 4:
+                newStyle = False
+            # but only for the C and C++ domains
+            if self.get_domain() and self.get_domain() not in ('c', 'cpp'):
+                newStyle = False
+            if newStyle:
+                # TODO: remove the "type: ignore" when Sphinx >= 4 is required
+                assert isinstance(n, addnodes.desc_sig_keyword)  # type: ignore
+                declarator[0] = addnodes.desc_sig_keyword(  # type: ignore
+                    display_obj_type, display_obj_type)
+            else:
+                assert isinstance(n, addnodes.desc_annotation)
+                assert n.astext()[-1] == " "
+                txt = display_obj_type + ' '
+                declarator[0] = addnodes.desc_annotation(txt, txt)
         if not self.app.env.config.breathe_debug_trace_doxygen_ids:
             target = self.create_doxygen_target(node)
         declarator.insert(0, target)
@@ -898,35 +912,12 @@ class SphinxRenderer:
         contentnode.extend(description)
         return nodes
 
-    def visit_unicode(self, node) -> List[Node]:
-        # Skip any nodes that are pure whitespace
-        # Probably need a better way to do this as currently we're only doing
-        # it skip whitespace between higher-level nodes, but this will also
-        # skip any pure whitespace entries in actual content nodes
-        #
-        # We counter that second issue slightly by allowing through single white spaces
-        #
-        if node.strip():
-            delimiter = None
-            if "<linebreak>" in node:
-                delimiter = "<linebreak>"
-            elif "\n" in node:
-                delimiter = "\n"
-            if delimiter:
-                # Render lines as paragraphs because RST doesn't have line breaks.
-                return [nodes.paragraph('', '', nodes.Text(line))
-                        for line in node.split(delimiter) if line.strip()]
-            return [nodes.Text(node)]
-        if node == six.u(" "):
-            return [nodes.Text(node)]
-        return []
-
     def visit_doxygen(self, node) -> List[Node]:
         nodelist = []
 
         # Process all the compound children
-        for compound in node.get_compound():
-            nodelist.extend(self.render(compound))
+        for n in node.get_compound():
+            nodelist.extend(self.render(n))
         return nodelist
 
     def visit_doxygendef(self, node) -> List[Node]:
@@ -1162,6 +1153,7 @@ class SphinxRenderer:
 
             rst_node.document = self.state.document
             rst_node['objtype'] = kind
+            rst_node['domain'] = self.get_domain() if self.get_domain() else 'cpp'
 
             contentnode = addnodes.desc_content()
             rst_node.append(contentnode)
@@ -1372,7 +1364,18 @@ class SphinxRenderer:
         if self.context and self.context.directive_args[0] == "doxygenpage":
             nodelist.extend(self.render_iterable(node.ordered_children))
         else:
-            nodelist.extend(self.render_iterable(node.content))
+            contentNodeCands = self.render_iterable(node.content)
+            # if there are consecutive nodes.Text we should collapse them
+            # and rerender them to ensure the right paragraphifaction
+            contentNodes = []  # type: List[Node]
+            for n in contentNodeCands:
+                if len(contentNodes) != 0 and isinstance(contentNodes[-1], nodes.Text):
+                    if isinstance(n, nodes.Text):
+                        prev = contentNodes.pop()
+                        contentNodes.extend(self.render_string(prev.astext() + n.astext()))
+                        continue  # we have handled this node
+                contentNodes.append(n)
+            nodelist.extend(contentNodes)
             nodelist.extend(self.render_iterable(node.images))
 
             paramList = self.render_iterable(node.parameterlist)
@@ -1700,6 +1703,7 @@ class SphinxRenderer:
 
         descnode = addnodes.desc()
         descnode['objtype'] = 'xrefsect'
+        descnode['domain'] = self.get_domain() if self.get_domain() else 'cpp'
         descnode += signode
         descnode += contentnode
 
@@ -1710,6 +1714,7 @@ class SphinxRenderer:
         for varlistentry, listitem in zip(node.varlistentries, node.listitems):
             descnode = addnodes.desc()
             descnode['objtype'] = 'varentry'
+            descnode['domain'] = self.get_domain() if self.get_domain() else 'cpp'
             signode = addnodes.desc_signature()
             signode += self.render_optional(varlistentry)
             descnode += signode
@@ -1779,7 +1784,7 @@ class SphinxRenderer:
 
         return [table]
 
-    def visit_mixedcontainer(self, node) -> List[Node]:
+    def visit_mixedcontainer(self, node: compoundsuper.MixedContainer) -> List[Node]:
         return self.render_optional(node.getValue())
 
     def visit_description(self, node) -> List[Node]:
@@ -1995,6 +2000,7 @@ class SphinxRenderer:
 
         desc = addnodes.desc()
         desc['objtype'] = 'friendclass'
+        desc['domain'] = self.get_domain() if self.get_domain() else 'cpp'
         signode = addnodes.desc_signature()
         desc += signode
 
@@ -2013,14 +2019,15 @@ class SphinxRenderer:
         signode += nodes.Text(node.name)
         return [desc]
 
-    def visit_param(self, node) -> List[Node]:
+    def visit_param(self, node: compound.paramTypeSub, *,
+                    insertDeclNameByParsing: bool = False) -> List[Node]:
         nodelist = []
 
         # Parameter type
         if node.type_:
             type_nodes = self.render(node.type_)
             # Render keywords as annotations for consistency with the cpp domain.
-            if len(type_nodes) > 0 and isinstance(type_nodes[0], six.text_type):
+            if len(type_nodes) > 0 and isinstance(type_nodes[0], str):
                 first_node = type_nodes[0]
                 for keyword in ['typename', 'class']:
                     if first_node.startswith(keyword + ' '):
@@ -2031,10 +2038,30 @@ class SphinxRenderer:
 
         # Parameter name
         if node.declname:
-            if nodelist:
-                nodelist.append(nodes.Text(" "))
-            nodelist.append(nodes.emphasis(text=node.declname))
+            dom = self.get_domain()
+            if not dom:
+                dom = 'cpp'
+            appendDeclName = True
+            if insertDeclNameByParsing:
+                if dom == 'cpp' and sphinx.version_info >= (4, 1, 0):
+                    parser = cpp.DefinitionParser(
+                        ''.join(n.astext() for n in nodelist),
+                        location=self.state.state_machine.get_source_and_line(),
+                        config=self.app.config)
+                    ast = parser._parse_type(named='single', outer='templateParam')
+                    assert ast.name is None
+                    nn = cpp.ASTNestedName(
+                        names=[cpp.ASTNestedNameElement(cpp.ASTIdentifier(node.declname), None)],
+                        templates=[False], rooted=False)
+                    ast.name = nn
+                    # the actual nodes don't matter, as it is astext()-ed later
+                    nodelist = [nodes.Text(str(ast))]
+                    appendDeclName = False
 
+            if appendDeclName:
+                if nodelist:
+                    nodelist.append(nodes.Text(" "))
+                nodelist.append(nodes.emphasis(text=node.declname))
         elif self.output_defname and node.defname:
             # We only want to output the definition name (from the cpp file) if the declaration name
             # (from header file) isn't present
@@ -2053,6 +2080,16 @@ class SphinxRenderer:
 
         return nodelist
 
+    def visit_templateparamlist(self, node: compound.templateparamlistTypeSub) -> List[Node]:
+        nodelist = []  # type: List[Node]
+        self.output_defname = False
+        for i, item in enumerate(node.param):
+            if i:
+                nodelist.append(nodes.Text(", "))
+            nodelist.extend(self.visit_param(item, insertDeclNameByParsing=True))
+        self.output_defname = True
+        return nodelist
+
     def visit_docparamlist(self, node) -> List[Node]:
         """Parameter/Exception/TemplateParameter documentation"""
 
@@ -2066,31 +2103,44 @@ class SphinxRenderer:
         # https://docutils.sourceforge.io/docs/ref/doctree.html#field-list
         fieldList = nodes.field_list()
         for item in node.parameteritem:
+            # TODO: does item.parameternamelist really have more than 1 parametername?
+            assert len(item.parameternamelist) <= 1, item.parameternamelist
+            nameNodes = []
+            parameterDirectionNodes = []
+            if len(item.parameternamelist) != 0:
+                paramNameNodes = item.parameternamelist[0].parametername
+                # TODO: how many elements can paramNameNodes have?
+                assert len(paramNameNodes) <= 1, paramNameNodes
+                if len(paramNameNodes) != 0:
+                    content = paramNameNodes[0].content_
+                    # this is really a list of MixedContainer objects, i.e., a generic object
+                    # we assume there is either 1 or 2 elements, if there is 2 the first is the
+                    # parameter direction
+                    assert len(content) == 1 or len(content) == 2, content
+                    nameNodes = self.render(content[-1])
+                    if len(content) == 2:
+                        dir = ''.join(n.astext() for n in self.render(content[0])).strip()
+                        assert dir in ('[in]', '[out]', '[inout]'), ">" + dir + "<"
+                        parameterDirectionNodes = [
+                            nodes.strong(dir, dir),
+                            nodes.Text(' ', ' ')
+                        ]
+
             name = nodes.field_name(
                 '', nodes.Text(fieldListName[node.kind] + ' '),
-                *self.render_iterable(item.parameternamelist))
-            body = nodes.field_body('', *self.render_optional(item.parameterdescription))
+                *nameNodes)
+            bodyNodes = self.render_optional(item.parameterdescription)
+            # TODO: is it correct that bodyNodes is either empty or a single paragraph?
+            assert len(bodyNodes) <= 1, bodyNodes
+            if len(bodyNodes) == 1:
+                assert isinstance(bodyNodes[0], nodes.paragraph)
+                bodyNodes = [nodes.paragraph(
+                    '', '', *(parameterDirectionNodes + bodyNodes[0].children))]
+            body = nodes.field_body('', *bodyNodes)
             field = nodes.field('', name, body)
             fieldList += field
         return [fieldList]
 
-    def visit_docparamnamelist(self, node) -> List[Node]:
-        """ Parameter Name Renderer"""
-        return self.render_iterable(node.parametername)
-
-    def visit_docparamname(self, node) -> List[Node]:
-        return self.render_iterable(node.content_)
-
-    def visit_templateparamlist(self, node) -> List[Node]:
-        nodelist = []  # type: List[Node]
-        self.output_defname = False
-        for i, item in enumerate(node.param):
-            if i:
-                nodelist.append(nodes.Text(", "))
-            nodelist.extend(self.render(item))
-        self.output_defname = True
-        return nodelist
-
     def visit_unknown(self, node) -> List[Node]:
         """Visit a node of unknown type."""
         return []
@@ -2162,10 +2212,8 @@ class SphinxRenderer:
         "mixedcontainer": visit_mixedcontainer,
         "description": visit_description,
         "param": visit_param,
-        "docparamlist": visit_docparamlist,
-        "docparamnamelist": visit_docparamnamelist,
-        "docparamname": visit_docparamname,
         "templateparamlist": visit_templateparamlist,
+        "docparamlist": visit_docparamlist,
         "docxrefsect": visit_docxrefsect,
         "docvariablelist": visit_docvariablelist,
         "docvarlistentry": visit_docvarlistentry,
@@ -2175,7 +2223,33 @@ class SphinxRenderer:
         "docentry": visit_docentry,
     }
 
-    def render(self, node, context: Optional[RenderContext] = None) -> List[Node]:
+    def render_string(self, node: str) -> List[Union[nodes.Text, nodes.paragraph]]:
+        # Skip any nodes that are pure whitespace
+        # Probably need a better way to do this as currently we're only doing
+        # it skip whitespace between higher-level nodes, but this will also
+        # skip any pure whitespace entries in actual content nodes
+        #
+        # We counter that second issue slightly by allowing through single white spaces
+        #
+        stripped = node.strip()
+        if stripped:
+            delimiter = None
+            if "<linebreak>" in stripped:
+                delimiter = "<linebreak>"
+            elif "\n" in stripped:
+                delimiter = "\n"
+            if delimiter:
+                # Render lines as paragraphs because RST doesn't have line breaks.
+                return [nodes.paragraph('', '', nodes.Text(line.strip()))
+                        for line in node.split(delimiter) if line.strip()]
+            # importantly, don't strip whitespace as visit_docpara uses it to collapse
+            # consecutive nodes.Text and rerender them with this function.
+            return [nodes.Text(node)]
+        if node == " ":
+            return [nodes.Text(node)]
+        return []
+
+    def render(self, node: Node, context: Optional[RenderContext] = None) -> List[Node]:
         if context is None:
             self.context = cast(RenderContext, self.context)
             context = self.context.create_child_context(node)
@@ -2184,18 +2258,18 @@ class SphinxRenderer:
             self.context = cast(RenderContext, self.context)
             if not self.filter_.allow(self.context.node_stack):
                 pass
-            elif isinstance(node, six.string_types):
-                result = self.visit_unicode(node)
+            elif isinstance(node, str):
+                result = self.render_string(node)
             else:
                 method = SphinxRenderer.methods.get(node.node_type, SphinxRenderer.visit_unknown)
-                result = method(self, node)
+                result = method(self, node)  # type: ignore
         return result
 
-    def render_optional(self, node) -> List[Node]:
+    def render_optional(self, node: Node) -> List[Node]:
         """Render a node that can be None."""
         return self.render(node) if node else []
 
-    def render_iterable(self, iterable) -> List[Node]:
+    def render_iterable(self, iterable: List[Node]) -> List[Node]:
         output = []
         for entry in iterable:
             output.extend(self.render(entry))
